from ast import literal_eval
from enum import Enum
from typing import (
    Any,
    Callable,
    List,
    Literal,
    MutableSequence,
    Optional,
    Sequence,
)

from openai import OpenAI
from openai.types.chat import ChatCompletion, ChatCompletionMessageToolCall
from pydantic import PrivateAttr

from light_agents.config import appSettings
from light_agents.core.logger_config import setup_logger
from light_agents.core.tool_registry import ToolRegistry
from light_agents.exceptions.thread_agent_exceptions import (
    AIAgentCompletionError,
)
from light_agents.schemas import ToolBaseSchema
from light_agents.schemas.messages_schemas import (
    Message,
    MessageBase,
    MessageRole,
    MessageType,
    ToolUseMessage,
)
from light_agents.schemas.thread_agent_schema import ThreadAgent
from light_agents.serializers.messages.openai_messages_serializer import (
    openai_messages_list_serializer,
)
from light_agents.serializers.tools.openai_tools_serializer import (
    openai_tool_calling_serializer,
)

logger = setup_logger(__name__)

client = OpenAI(api_key=appSettings.OPENAI_API_KEY)


class OpenAIMessageRoles(str, Enum):
    """OpenAI Message Roles."""

    USER = "user"
    AI = "assistant"
    SYSTEM = "system"

    def get_role_mapping(self) -> dict[str, str]:
        """Get role mapping from OpenAI roles to Light Agent roles."""
        role_mapping = {
            MessageRole.AI.value: OpenAIMessageRoles.AI.value,
            MessageRole.USER.value: OpenAIMessageRoles.USER.value,
            MessageRole.SYSTEM.value: OpenAIMessageRoles.SYSTEM.value,
        }
        return role_mapping


class OpenAIAgent(ThreadAgent):
    """OpenAI Agent.

    Interacts with OpenAI model's API.

    Attributes:
        model: OpenAI model identifier. Default is `gpt-4o`.
        max_tokens: Maximum number of tokens to generate.
        messages_serializer: Messages serializer function for the model.
        verbose: Flag to indicate if the agent should print messages.
        tools: List of tools available for the agent.
        tools_serializer: Tools serializer function for the model.
        tools_registry: Registry of tools available for the agent.
        system_message_selector: Selector for system messages.

    """

    model: str = "gpt-4o"
    """OpenAI model identifier. Default is `gpt-4o`."""

    max_tokens: int = 8192
    """Maximum number of tokens to generate."""

    messages_serializer: Callable[..., Any] = openai_messages_list_serializer
    """Messages serializer function for the model."""

    verbose: bool = True
    """Flag to indicate if the agent should print messages."""

    tools: List[ToolBaseSchema] = []
    """List of tools available for the agent."""

    tools_serializer: Callable[..., Any] = openai_tool_calling_serializer
    """Tools serializer function for the model."""

    tools_registry: Optional[ToolRegistry] = None
    """Registry of tools available for the agent. at
    [light_agents.core.tool_registry.ToolRegistry]"""

    system_message_selector: Literal["first", "last", "all"] = "all"
    """Selector for system messages."""

    _current_run_messages: List[MessageBase] = PrivateAttr(default=[])
    """List of messages generated in the current run."""

    def __init__(self, **data: Any) -> None:
        """Initialize OpenAI Agent."""
        super().__init__(**data)
        self.tools_registry = ToolRegistry()
        if len(self.tools) > 0:
            self.tools_registry.register_tools(self.tools)

    def agent_run(
        self, thread_messages: MutableSequence[MessageBase], **kwargs: Any
    ) -> MutableSequence[MessageBase]:
        """Run the agent on the thread messages.

        The process of running the agent is separated in the following steps:
        1. Send the messages to the model in `send_to_openai`. Retrieve model
            run response.

        2. Process the model response in `process_model_response`. If the
            response is a simple text message, `_current_run_messages` will be
            a list of `Message` objects.

        3. Process the tools in `process_tools`. Execute all required tools.
            `_current_run_messages` will be a list containing `ToolUseMessage`
            objects with the `tool_outputs` field populated.

        4. If there is any `ToolUseMessage` in the `_current_run_messages`,
            call `agent_run` again with the updated messages in order to get a
            new response from the model based on the tools outputs.

        Args:
            thread_messages: List of messages in the thread.
            **kwargs: Additional arguments.

        Returns:
           messages: List of messages generated by the agent.

        Examples:
        >>> agent = Agent()
        >>> agent.agent_run([Message(content="Hello",type="text", role="user)])
        [Message(content="This is a Response", role="ai", type="text")]

        """
        if not kwargs.get("calling_from_inside_agent_run"):
            self._current_run_messages = []

        logger.debug("Running OpenAI Agent.") if self.verbose else None

        model_response = self.send_to_openai(thread_messages, **kwargs)
        logger.debug(
            f"Model response: {model_response}\nProcessing it"
        ) if self.verbose else None

        run_messages = self.process_model_response(model_response, **kwargs)
        self._current_run_messages.extend(run_messages)
        thread_messages.extend(run_messages)
        logger.debug(
            f"Model response processed: {run_messages}"
        ) if self.verbose else None

        if any(
            isinstance(message, ToolUseMessage) for message in run_messages
        ):
            kwargs["calling_from_inside_agent_run"] = True
            logger.debug(
                "Tools were used in the response. Processing tools."
            ) if self.verbose else None
            self.agent_run(thread_messages, **kwargs)

        return self._current_run_messages

    def send_to_openai(
        self, thread_messages: MutableSequence[MessageBase], **kwargs: Any
    ) -> ChatCompletion:
        """Send messages to OpenAI model.

        Args:
            thread_messages: List of messages in the thread.
            **kwargs: Additional arguments.

        Returns:
            completion: OpenAI completion object.

        """
        messages = self.messages_serializer(thread_messages)
        completion: ChatCompletion
        if len(self.tools) > 0:
            logger.debug("Calling agent with tools.") if self.verbose else None
            serialized_tools = [
                self.tools_serializer(tool) for tool in self.tools
            ]
            logger.debug(
                f"Serialized tools: {serialized_tools}"
            ) if self.verbose else None
            completion = client.chat.completions.create(
                model=self.model,
                messages=messages,
                max_tokens=self.max_tokens,
                tools=serialized_tools,
            )
        else:
            completion = client.chat.completions.create(
                model=self.model,
                messages=messages,
                max_tokens=self.max_tokens,
            )
        return completion

    def process_model_response(
        self, completion: ChatCompletion, **kwargs: Any
    ) -> Sequence[MessageBase]:
        """Process the model response.

        If the response contains a tool calling, calls `execute_tool` method.
        Otherwise returns the first generated message.

        Args:
            completion: OpenAI completion object.
            **kwargs: Additional arguments.

        Returns:
            messages: List of messages generated by the agent.

        """
        response_messages = []
        # tool_responses = []
        for choice in completion.choices:
            if choice.finish_reason == "stop":
                logger.debug(
                    "Model stopped by it's own."
                ) if self.verbose else None
                # natural stop for the model
                if choice.message.content:
                    response_content: str = choice.message.content
                    response_message = Message(
                        content=response_content,
                        role=MessageRole.AI,
                        type=MessageType.TEXT,
                    )
                    response_messages.append(response_message)
                    return response_messages
                
                else:
                    raise AIAgentCompletionError(
                        "Model stopped without generating any content. "
                        f"Response: \n{choice}"
                    )

            elif choice.finish_reason in ["tool_calls", "function_call"]:
                logger.debug(
                    f"Tool call detected: {choice.finish_reason}"
                ) if self.verbose else None

                lightagents_tool_use_messages = []
                openai_tool_calls_messages = choice.message.tool_calls
                if (
                    openai_tool_calls_messages
                    and len(openai_tool_calls_messages) > 0
                ):
                    for tool_call in openai_tool_calls_messages:
                        if not isinstance(
                            tool_call, ChatCompletionMessageToolCall
                        ):
                            raise AIAgentCompletionError(
                                "Unknown message type for tool call. Received "
                                f"'{type(choice)}', expected "
                                "'ChatCompletionMessageToolCall'."
                            )

                        tool_use_message = ToolUseMessage(
                            run_id=tool_call.id,
                            role=MessageRole.TOOL_USE,
                            type=MessageType.TEXT,
                            external_fields=kwargs,
                            name=tool_call.function.name,
                            input_params_dict=tool_call.function.arguments,
                        )
                        lightagents_tool_use_messages.append(tool_use_message)

                    lightagents_tool_use_messages = self.process_tools(
                        lightagents_tool_use_messages, **kwargs
                    )

                    return lightagents_tool_use_messages

                else:
                    raise AIAgentCompletionError(
                        "Model completion contained a tool call, but no tool "
                        "calls were found."
                    )

            elif choice.finish_reason == "length":
                logger.warning("Model reached maximum token limits.")
                # TODO: handle this cas
                ...
                
                return []

            elif choice.finish_reason == "content_filter":
                # TODO: handle this case
                logger.warning("Model content omitted due to content filter.")
                ...
                return []

            else:
                logger.error(
                    f"Unknown finish reason: '{choice.finish_reason}'"
                )
                return []

        raise AIAgentCompletionError(f"No messages were\n{completion}")
        

    def process_tools(
        self, tool_use_messages: List[ToolUseMessage], **kwargs: Any
    ) -> List[ToolUseMessage]:
        """Process tool use messages.

        Args:
            tool_use_messages: List of tool use messages.
            **kwargs: Additional arguments.

        Returns:
            tool_use_messages: List of tool use messages.

        """
        updated_tool_use_messages = []
        if self.tools_registry:
            for tool_message in tool_use_messages:
                if isinstance(tool_message.input_params_dict, str):
                    try:
                        args_dict = literal_eval(
                            tool_message.input_params_dict
                        )
                    except ValueError:
                        logger.warning(
                            "Could not parse input args. Using as str."
                        )
                        args_dict = {
                            "response": tool_message.input_params_dict
                        }
                else:
                    args_dict = tool_message.input_params_dict

                try:
                    logger.debug(
                        f"Executing tool: '{tool_message.name}' "
                        f"with args: {args_dict}"
                    ) if self.verbose else None

                    tool_response = self.tools_registry.execute_tool(
                        tool_message.name, args_dict, **kwargs
                    )

                    logger.info(
                        f"Tool returned: \n{tool_response}"
                    ) if self.verbose else None

                    tool_message.tool_outputs = tool_response.content
                    if tool_response.external_fields:
                        tool_message.external_fields.update(
                            tool_response.external_fields
                        )
                    updated_tool_use_messages.append(tool_message)
                    
                except Exception as e:
                    tool_message.is_error = True
                    logger.error(
                        f"Error executing tool: '{tool_message.name}'"
                    )
                    raise ValueError(f"Error executing tool: {e}")
                
        return updated_tool_use_messages



from enum import Enum
from typing import Any, Dict, MutableSequence, Sequence

from pydantic import BaseModel

from light_agents.schemas.messages_schemas import MessageBase
from light_agents.schemas.model_config import model_config
from light_agents.schemas.thread_agent_schema import ThreadAgent


class ThreadType(str, Enum):
    """Possible types for a single Thread."""

    BASIC = "basic"


class ThreadBase(BaseModel):
    """Base schema for a Thread."""

    model_config = model_config
    type: ThreadType
    messages: MutableSequence[MessageBase] = []
    external_thread_fields: Dict[str, Any] = {}

    def add_message(self, message: MessageBase) -> None:
        """Add a message to the thread."""
        if not isinstance(message, MessageBase):
            raise ValueError("The message must be an instance of MessageBase.")

        self.messages.append(message)

    def add_messages_list(
        self, messages: Sequence[MessageBase]
    ) -> None:
        """Add a list of messages to the thread."""
        if not all(isinstance(message, MessageBase) for message in messages):
            raise ValueError("All messages must be instances of MessageBase.")

        for message in messages:
            self.add_message(message)

    def process_thread(self, thread_agent: ThreadAgent) -> None:
        """Process the thread."""
        if not isinstance(thread_agent, ThreadAgent):
            raise ValueError(
                "The thread agent must be an instance of ThreadAgent."
            )

        ## TODO:
        ## 1. Deal gracefully with exceptions
        ## 2. Update external thread fields based on the agent's output
        messages = thread_agent.agent_run(
            self.messages, **self.external_thread_fields
        )
        self.add_messages_list(messages)



from abc import ABC, abstractmethod
from typing import Any, List, Optional

from pydantic import BaseModel, ConfigDict, Field


class ToolBaseSchema(ABC, BaseModel):
    """Base schema for a Tool used in LLM."""

    model_config = ConfigDict(extra="allow")
    name: str = Field(..., description="The name of the function")
    description: str = Field(
        ..., description="A brief description of what the function does"
    )
    json_response: Optional[bool] = Field(
        default=False, description="Flag to indicate if the response is a JSON"
    )
    required: Optional[List[str]] = Field(default=[], description="The required fields")

    @abstractmethod
    def run(self, *args: Any, **kwargs: Any) -> "ToolResponseSchema":
        """Python  function to be executed when the tool is called."""
        pass


class ToolResponseSchema(BaseModel):
    """Base schema for a Tool response used by an LLM."""

    content: str
    """String content to be returned by the tool to the LLM.
     
     It can include any content or extra prompt to guide the LLM based on the
     tool's output.
     """

    external_fields: Optional[dict[str, Any]] = {}
    """External fields to be returned for the thread.
     
     Returns extra information and update thread fields based on
     the tool's output, without feeding the LLM with the content.
     Usefull for passing sensitive information to prevent prompt injection. 
     """

    is_error: Optional[bool] = False
    """Flag to indicate if the tool response is an error."""



import logging
from typing import Any, Dict, Literal, Optional

from termcolor import colored

Color = Literal[
    "black",
    "grey",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "light_grey",
    "dark_grey",
    "light_red",
    "light_green",
    "light_yellow",
    "light_blue",
    "light_magenta",
    "light_cyan",
    "white",
]


class ColoredFormatter(logging.Formatter):
    """Custom log formatter that colorizes the entire log message."""

    def format(self, record: logging.LogRecord) -> str:
        """Format the log record."""
        colors: Dict[str, Color] = {
            "DEBUG": "cyan",
            "INFO": "green",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "magenta",
        }

        levelname = record.levelname
        if levelname in colors:
            color = colors[levelname]
            message = super(ColoredFormatter, self).format(record)
            colored_message: str = colored(message, color)
            return colored_message

        return super(ColoredFormatter, self).format(record)


def setup_logger(name: Optional[str] = None, **kwargs: Any) -> logging.Logger:
    """Set up the logger."""
    log_level = kwargs.get("log_level", logging.DEBUG)

    logger = logging.getLogger(name)
    logger.setLevel(log_level)

    if not logger.handlers:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.DEBUG)

        formatter = ColoredFormatter(
            "%(asctime)s [%(levelname)s] %(name)s:\n%(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )

        # Adicionar o formatter ao handler
        console_handler.setFormatter(formatter)

        # Adicionar o handler ao logger
        logger.addHandler(console_handler)

        # Opcional: Propagar o log para loggers de nível superior
        logger.propagate = False

    return logger



class AIAgentCompletionError(Exception):
    """Exception raised when an AI agent fails to complete a task."""

    def __init__(self, message: str) -> None:
        """Initialize the exception."""
        super().__init__(message)
        self.message = message



from enum import Enum
from typing import Any


def python_type_to_json_type(value: Any) -> str:
    """Convert a Python type to a JSON type."""
    if isinstance(value, str):
        return "string"
    elif isinstance(value, (int, float)):
        return "number"
    elif isinstance(value, bool):
        return "boolean"
    elif isinstance(value, (list, tuple)):
        return "array"
    elif isinstance(value, dict):
        return "object"
    elif isinstance(value, Enum):
        return "enum"
    elif value is None:
        return "null"
    else:
        return "string"



from enum import Enum
from typing import Any, Dict

from light_agents.core.logger_config import setup_logger
from light_agents.schemas import ToolBaseSchema, ToolUseMessage
from light_agents.serializers.tools.base_serializers import (
    python_type_to_json_type,
)

logger = setup_logger(__name__)


def openai_tool_calling_serializer(
    llm_function: ToolBaseSchema,
) -> Dict[str, Any]:
    """Serialize a single `ToolBaseSchema` into OpenAI API format.

    OpenAI tool schema looks like this:
    ```json
        {
            "type": "function",
            "function": {
                "name": "get_delivery_date",
                "description": "Get the delivery date for a customer's order.",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "order_id": {
                            "type": "string",
                            "description": "The customer's order ID.",
                        },
                    },
                    "required": ["order_id"],
                    "additionalProperties": false,
                },
            }
    }
    ```
    """
    # TODO: Implemment JSON structured mode.
    function_dict = llm_function.model_dump(
        exclude={"name", "description", "required", "json_response"}
    )

    serialized_tool: Dict[str, Any] = {
        "type": "function",
        "function": {
            "name": llm_function.name,
            "description": llm_function.description,
            "parameters": {
                "type": "object",
                "properties": {},
            },
        },
    }

    if llm_function.required and len(llm_function.required) > 0:
        serialized_tool["function"]["parameters"]["required"] = (
            llm_function.required
        )

    for param in function_dict.keys():
        json_param_type = python_type_to_json_type(function_dict[param])
        serialized_tool["function"]["parameters"]["properties"][param] = {
            "type": json_param_type,
            "description": llm_function.model_fields[param].description,
        }

        annotation = llm_function.model_fields[param].annotation
        if (
            annotation
            and isinstance(annotation, type)
            and issubclass(annotation, Enum)
        ):
            serialized_tool["function"]["parameters"]["properties"][param][
                "enum"
            ] = list(annotation.__members__.keys())

    return serialized_tool


def openai_tooL_response_serializer(
    tool: ToolUseMessage, **kwargs: Any
) -> list[Dict[str, Any]]:
    """Serialize a single tool response for the OpenAI API.

    The function creates both ```assistant``` tool request block and ```user```
    ```json
    [
        {
            "role": "assistant",
            "tool_calls": [
                {
                    "id": "call_xxxx",
                    "type": "function",
                    "function": {
                        "arguments": {"key1": "value1", "key2": "value2"},
                        "name": "tool_name",
                    }
                }
            ]
        },
        {
            "role": "tool",
            "content": "tool_response_content",
            "tool_call_id": "call_xxxx"
        }
    ]
    ```
    """
    if not isinstance(tool, ToolUseMessage):
        raise ValueError("tool must be an instance of ToolUseMessage.")

    tool_output = tool.tool_outputs
    # TODO: gracefully handle non-string tool_output
    if not isinstance(tool_output, str):
        logger.warning("Tool output is not a string. Converting it to string.")
        tool_output = str(tool_output)

    serialized_tool_calling_message = {
        "role": "assistant",
        "tool_calls": [
            {
                "id": tool.run_id,
                "type": "function",
                "function": {
                    "arguments": tool.input_params_dict,
                    "name": tool.name,
                },
            }
        ],
    }

    serialized_tool_response_message = {
        "role": "tool",
        "content": tool_output,
        "tool_call_id": tool.run_id,
    }

    return [serialized_tool_calling_message, serialized_tool_response_message]



from light_agents.serializers.tools.claude_tools_serializer import (
    claude_tool_calling_serializer,
)
from light_agents.serializers.tools.openai_tools_serializer import (
    openai_tool_calling_serializer,
)

__all__ = [
    "claude_tool_calling_serializer",
    "openai_tool_calling_serializer",
]



from typing import Any, List, MutableSequence

from light_agents.schemas.messages_schemas import (
    MessageBase,
)


def base_serializer(
    messages: MutableSequence[MessageBase], **kwargs: Any
) -> List[dict[str, str]]:
    """Serialize only messages role and content."""
    if not isinstance(messages, list):
        raise ValueError("The messages must be a list.")

    serialized_messages = []
    for message in messages:
        serialized_messages.append(
            {"content": message.content, "role": message.role.value}
        )
    return serialized_messages


def role_mapping_serializer(
    messages: MutableSequence[MessageBase], **kwargs: Any
) -> List[dict[str, str]]:
    """Serialize only messages role and content.

    It uses a mapping to change the roles to supported values for other LLMs.
    If a role is not in the mapping, it will be ignored.
    """
    if "roles_mapping" not in kwargs:
        raise ValueError("'roles_mapping' dict must be provided.")

    if not isinstance(messages, list):
        raise ValueError("The messages must be a list.")
    serialized_messages = []
    for message in messages:
        converted_role = kwargs["roles_mapping"].get(message.role, None)
        if converted_role:
            serialized_messages.append(
                {"content": message.content, "role": converted_role}
            )
    return serialized_messages



from typing import Any


class MessageSupportException(ValueError):
    """Exception raised for unsupported messages.

    Usefull for every error that involves message types or roles.
    """
    
    def __init__(self, message: str, *args: Any, **kwargs: Any) -> None:
        """Initialize the exception."""
        super().__init__(message)


class MessageTypeNotSupportedException(MessageSupportException):
    """Exception raised for unsupported message types."""

    def __init__(self, message_type: str, supported_types: list[Any]) -> None:
        """Initialize the exception."""
        self.message_type = message_type
        self.supported_types = supported_types
        super().__init__(
            f"Message type '{message_type}' is not supported. "
            f"Supported types are: {supported_types}"
        )


class MessageRoleNotSupportedException(MessageSupportException):
    """Exception raised for unsupported message roles."""

    def __init__(self, message_role: str, supported_roles: list[Any]) -> None:
        """Initialize the exception."""
        self.message_role = message_role
        self.supported_roles = supported_roles
        super().__init__(
            f"Message role '{message_role}' is not supported. "
            f"Supported roles are: {supported_roles}"
        )



from typing import Any, Dict, List, MutableSequence

from light_agents.core.logger_config import setup_logger
from light_agents.exceptions.messages_exceptions import (
    MessageRoleNotSupportedException,
    MessageSupportException,
    MessageTypeNotSupportedException,
)
from light_agents.schemas.messages_schemas import (
    Message,
    MessageBase,
    MessageRole,
    MessageType,
    ToolUseMessage,
)
from light_agents.serializers.tools.openai_tools_serializer import (
    openai_tooL_response_serializer,
)

logger = setup_logger(__name__)


def openai_text_message_serializer(
    text_message: Message, **kwargs: Any
) -> Dict[str, Any]:
    """Serialize a single text message to OpenAI's API format.

    Warning:
        Handles only text messages.

    The openAI serialized message is a json:
    ```json
    {
        "role": "user/assistant/system",
        "content": [
            {
                "type": "text",
                "text": "Content of the message"
            }
        ]
    }
    ```

    Arguments:
        text_message: The message to be serialized.
        **kwargs: Additional arguments.

    Returns:
        serialized_message: The serialized message

    Examples:
        >>> from light_agents.schemas.messages_schemas import Message, MessageType, MessageRole
        >>> message = Message(
        ...     role=MessageRole.USER,
        ...     type=MessageType.TEXT,
        ...     content="Hello, how are you?"
        ... )
        >>> openai_text_message_serializer(message)
        {
            "role": "user",
            "content": [
                {
                    "type": "text",
                    "text": "Hello, how are you?"
                }
            ]
        }

    """  # noqa: E501
    if not isinstance(text_message, Message):
        raise MessageSupportException("Only Message type is supported.")
    if text_message.type != MessageType.TEXT:
        raise MessageTypeNotSupportedException(
            text_message.type, [MessageType.TEXT]
        )

    role_mapping = {
        MessageRole.USER.value: "user",
        MessageRole.AI.value: "assistant",
        MessageRole.SYSTEM.value: "system",
    }
    if "roles_mapping" in kwargs:
        role_mapping = kwargs["roles_mapping"]
    
    message_converted_role = role_mapping.get(text_message.role)
    supported_roles = ["user", "assistant", "system"]
    if message_converted_role not in supported_roles:
        raise MessageRoleNotSupportedException(
            text_message.role, supported_roles
        )

    serialized_message = {
        "role": message_converted_role,
        "content": [
            {
                "type": "text",
                "text": text_message.content,
            }
        ],
    }

    return serialized_message


def openai_messages_list_serializer(
    messages: MutableSequence[MessageBase], **kwargs: Any
) -> List[Dict[str, Any]]:
    """Serialize a series of messages for OpenAI API.

    Handles text messages only.
    """
    serialized_messages = []
    for message in messages:
        if isinstance(message, Message):
            if message.type == MessageType.TEXT:
                serialized_message = openai_text_message_serializer(message)
                serialized_messages.append(serialized_message)
            else:
                logger.warning(
                    f"'{message.type}' message type is not supported."
                    "Message will be ignored."
                )
                
        elif isinstance(message, ToolUseMessage):
            tool_serialized_messages = openai_tooL_response_serializer(message)
            if any(tool for tool in tool_serialized_messages):
                serialized_messages.extend(tool_serialized_messages)
            
        else:
            logger.warning(
                f"'{message}' message type is not yet supported."
                "Message will be ignored."
            )
    return serialized_messages



from light_agents.schemas.messages_schemas import Message, ToolUseMessage
from light_agents.schemas.tool_schema import ToolBaseSchema, ToolResponseSchema

__all__ = [
    "ToolBaseSchema",
    "ToolResponseSchema",
    "Message",
    "ToolUseMessage",
]



from typing import Any, Dict, List, MutableSequence

from light_agents.core.logger_config import setup_logger
from light_agents.schemas.messages_schemas import (
    Message,
    MessageBase,
    MessageRole,
    MessageType,
    ToolUseMessage,
)
from light_agents.serializers.tools.claude_tools_serializer import (
    claude_tool_response_serializer,
)

logger = setup_logger(__name__)


def claude_text_message_serializer(
    message: Message, **kwargs: Any
) -> Dict[str, Any] | None:
    """Seriliaze a single message for Claude API.

    Handles only text messages.

    Expected return:
    ```json
    {
        "role": "user/assistant",
        "content": [
            {
                "type": "text",
                "text": "message content"
            }
        ]
    }
    ```

    *It works for default claude SDK. Not tested with Bedrock Claude.*

    """
    role_mapping = {
        MessageRole.USER.value: "user",
        MessageRole.AI.value: "assistant",
    }
    if "roles_mapping" in kwargs:
        role_mapping = kwargs["roles_mapping"]

    if message.type != MessageType.TEXT:
        logger.warning(
            f"'{message.type}' message type is not supported. Message ignored."
        )
        return None

    if not role_mapping.get(message.role):
        logger.warning(f"'{message.role}' role is not supported. Ignoring message.")
        return None

    serialized_message = {
        "role": role_mapping.get(message.role),
        "content": message.content,
    }

    return serialized_message


def claude_messages_list_serializer(
    messages: MutableSequence[MessageBase], **kwargs: Any
) -> List[Dict[str, Any]]:
    """Serialize a series of messages for Claude API.

    Handles text messages and tool use messages.
    *Not tested with Bedrock Claude.*
    """
    serialized_messages = []
    for message in messages:
        if isinstance(message, Message):
            if message.type == MessageType.TEXT:
                ##TODO: handle text messages
                serialized_message = claude_text_message_serializer(message, **kwargs)
                if serialized_message:
                    serialized_messages.append(serialized_message)
            else:
                logger.warning(
                    f"'{message.type}' message type is not supported."
                    "Message will be ignored."
                )

        elif isinstance(message, ToolUseMessage):
            tool_serialized_messages = claude_tool_response_serializer(message)
            if any(tool for tool in tool_serialized_messages):
                serialized_messages.extend(tool_serialized_messages)

        else:
            logger.warning(
                f"'{type(message)}' message type is not supported."
                "Message will be ignored."
            )

    return serialized_messages


def bedrock_claude_messages_serializer(
    messages: MutableSequence[MessageBase], **kwargs: Any
) -> List[Dict[str, str]]:
    """Serialize messages for Bedrock Claude model.

    Return example:
    ```json
    [
        {
            "role": "user",
            "content": [
                {
                    "type": "text",
                    "text": "message content
                }
            ]
        }
    ]
    ```

    0. If a roles_mapping is provided, it will be used for role mapping.
    1. The roles must alternate between user and assistant.
        If there is two user messages inMutableSequence, they will be
        included inside "content" list.
    2. If there is any, only the first system message will be considered.
    """
    if not isinstance(messages, list):
        raise ValueError("The messages must be a list.")

    roles_mapping = MessageRole.get_roles_mapping()
    if "roles_mapping" in kwargs:
        roles_mapping = kwargs["roles_mapping"]

    serialized_messages = []
    current_message: Dict[str, Any] = {"role": None, "content": []}
    for message in messages:
        # TODO: Handle when last message is tool use or media message
        if isinstance(message, Message):
            current_role = roles_mapping.get(message.role, None)
            if current_role:
                if current_role == current_message["role"]:
                    if message.type == MessageType.TEXT:
                        current_message["content"].append(
                            {"type": "text", "text": message.content}
                        )
                    else:
                        # TODO: Gracefully handle other message types.
                        print("Only text messages are supported for now.")
                else:
                    if current_message.get("role", None):
                        serialized_messages.append(current_message)
                    current_message = {
                        "role": current_role,
                        "content": [{"type": "text", "text": message.content}],
                    }
    # TODO: Handle when last message is tool use or media message
    if isinstance(message, Message):
        serialized_messages.append(current_message)
    return serialized_messages



from typing import Optional

from pydantic_settings import BaseSettings, SettingsConfigDict


class AppSettings(BaseSettings):
    """Base settings."""

    model_config = SettingsConfigDict(
        env_file=".env",
        extra="ignore",  # Change made here
    )
    GCP_PROJECT_ID: Optional[str] = None
    GCP_REGION: Optional[str] = None
    AWS_ACCESS_KEY: Optional[str] = None
    AWS_SECRET_KEY: Optional[str] = None
    OPENAI_API_KEY: Optional[str] = None


appSettings = AppSettings()



from datetime import datetime, timezone
from enum import Enum
from typing import Any, Optional, Union

from pydantic import BaseModel

from light_agents.schemas.model_config import model_config


class MessageRole(str, Enum):
    """Possible roles for a single message."""

    USER = "user"
    SYSTEM = "system"
    AI = "ai"
    TOOL_USE = "tool_use"

    @staticmethod
    def get_roles_mapping() -> dict[str, str]:
        """Get the role mapping."""
        return {
            MessageRole.USER.value: "user",
            MessageRole.SYSTEM.value: "system",
            MessageRole.AI.value: "ai",
        }


class MessageType(str, Enum):
    """Possible types for a single message."""

    TEXT = "text"
    IMAGE = "image"
    VIDEO = "video"
    AUDIO = "audio"
    FILE = "file"


class MessageBase(BaseModel):
    """A single message."""

    model_config = model_config
    role: MessageRole
    type: MessageType
    external_fields: dict[str, Any] = {}
    """Fields to be returned for the thread outside LLM."""


class Message(MessageBase):
    """A single message.

    Attributes
    ----------
        content: the content of the message to be passed to the LLM
        timestamp: the timestamp of message creation

    """

    model_config = model_config
    content: str
    timestamp: Optional[datetime] = datetime.now(timezone.utc)


class ToolUseMessage(MessageBase):
    """Message for a LLM function calling.

    Attributes
    ----------
        run_id: the run id of the tool
        name: the name of the tool
        input_params_dict: the input parameters of the tool
        tool_outputs: the output of the tool
        is_error: flag to indicate if the tool execution was an error

    """

    run_id: str
    name: str
    input_params_dict: Union[str, dict[str, Any]]
    tool_outputs: Optional[Any] = None
    is_error: Optional[bool] = False


class MediaMessage(Message):
    """A single media message."""

    media_url: str



from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, MutableSequence, Sequence

from pydantic import BaseModel

from light_agents.schemas.messages_schemas import MessageBase, ToolUseMessage
from light_agents.schemas.model_config import model_config
from light_agents.utils.serializers import base_serializer


class ThreadAgent(ABC, BaseModel):
    """Abstract schema for a Thread agent."""

    model_config = model_config
    messages_serializer: Callable[..., List[Dict[str, str]]] = base_serializer
    verbose: bool = False

    @abstractmethod
    def agent_run(
        self, thread_messages: MutableSequence[MessageBase], **kwargs: Any
    ) -> Sequence[MessageBase]:
        """Run the agent."""
        raise NotImplementedError

    @abstractmethod
    def process_tools(
        self, tool_use_messages: List[ToolUseMessage], **kwargs: Any
    ) -> MutableSequence[ToolUseMessage]:
        """Process the tools."""
        raise NotImplementedError



from typing import Any, Callable, Dict, List

from light_agents.core.logger_config import setup_logger
from light_agents.schemas.tool_schema import ToolBaseSchema, ToolResponseSchema

logger = setup_logger(__name__)


class ToolRegistry:
    """Registry for tools that can be used by AI agents.

    This registry converts tools defined in the Pydantic model into
    callables that can be used by AI agents.
    """

    def __init__(self) -> None:
        """Initialize the ToolRegistry class."""
        self.tools: Dict[str, Callable[..., ToolResponseSchema]] = {}

    def register(self, tool: ToolBaseSchema) -> None:
        """Register a tool using its name and run method."""
        self.tools[tool.name] = tool.run

    def register_tools(self, tools: List[ToolBaseSchema]) -> None:
        """Register a list of tools."""
        for tool in tools:
            self.register(tool)

    def execute_tool(
        self, tool_name: str, args: Dict[str, Any], **kwargs: Any
    ) -> ToolResponseSchema:
        """Execute a tool with arguments given inside a dict.

        If any kwargs are given and they overlap with args, the args will be
        updated with the kwargs.
        """
        if tool_name not in self.tools:
            # To prevent error, tells the LLM that the tool wasn't available
            return ToolResponseSchema(
                content="<INTERNAL> This tool was not available </INTERNAL>",
                is_error=True,
            )

        tool = self.tools[tool_name]
        args.update(kwargs)

        if kwargs.get("verbose"):
            logger.debug(f"Executing tool '{tool_name}' with args: {args}")

        try:
            result = tool(**args)

            if not isinstance(result, ToolResponseSchema):
                raise ValueError(
                    f"Tool '{tool_name}' should return a ToolResponseSchema"
                )

            return result
        except Exception as e:
            logger.error(f"Error executing tool '{tool_name}': {e}")
            raise ValueError(f"Error executing tool '{tool_name}': {e}")



from enum import Enum
from typing import (
    Any,
    Callable,
    List,
    Literal,
    MutableSequence,
    Optional,
    Sequence,
)

from anthropic import AnthropicVertex
from anthropic.types import Message as AnthropicMessage
from anthropic.types import TextBlock as AnthropicTextBlock
from pydantic import PrivateAttr

from light_agents.config import appSettings
from light_agents.core.logger_config import setup_logger
from light_agents.core.tool_registry import ToolRegistry
from light_agents.schemas import ToolBaseSchema
from light_agents.schemas.messages_schemas import (
    Message,
    MessageBase,
    MessageRole,
    MessageType,
    ToolUseMessage,
)
from light_agents.schemas.thread_agent_schema import ThreadAgent
from light_agents.serializers.messages.claude_messages_serializers import (
    claude_messages_list_serializer,
)
from light_agents.serializers.tools import claude_tool_calling_serializer

logger = setup_logger(__name__)

client = AnthropicVertex(
    project_id=appSettings.GCP_PROJECT_ID,  # type: ignore
    region=appSettings.GCP_REGION,  # type: ignore
)

MODEL_ID = "claude-3-5-sonnet@20240620"

# client = AnthropicBedrock(
#     aws_access_key=appSettings.AWS_ACCESS_KEY,
#     aws_secret_key=appSettings.AWS_SECRET_KEY,
# )
# MODEL_ID = "anthropic.claude-3-5-sonnet-20240620-v1:0"


class ModelMessageRoles(Enum):
    """Message roles for the model."""

    USER = "user"
    AI = "assistant"

    @staticmethod
    def get_role_mapping() -> dict[str, str]:
        """Get the role mapping."""
        return {
            MessageRole.USER.value: ModelMessageRoles.USER.value,
            MessageRole.AI.value: ModelMessageRoles.AI.value,
        }


class ClaudeAgent(ThreadAgent):
    """Claude agent responsible for interacting with the Anthropic model.

    If any system message is passed inside the messages, the system message
    will be picked based on ```system_message_method```. The rest will be
    ignored.

    Attributes:
        model: The model identifier to be used for processing messages.
        max_tokens: Maximum number of tokens to be generated.
        serializer: Function used to serialize conversation messages.
        verbose: Whether to log detailed information for debugging.
        tools: List of tools available for the agent to use.
        tools_serializer: Function used to serialize the tools information.
        system_message_method: ```first``` or ```last``` to choose the system
            message inside messages.

    """

    model: str = MODEL_ID
    max_tokens: int = 8192
    messages_serializer: Callable[..., Any] = claude_messages_list_serializer
    verbose: bool = True
    tools: List[ToolBaseSchema] = []
    tools_serializer: Callable[..., Any] = claude_tool_calling_serializer
    tools_registry: Optional[ToolRegistry] = None
    system_message_selector: Literal["first", "last"] = "first"
    _current_run_messages: List[MessageBase] = PrivateAttr(default=[])
    """List of messages generated by the agent during the current run."""

    def __init__(self, **data: Any) -> None:
        """Initialize the Claude agent."""
        super().__init__(**data)
        self.tools_registry = ToolRegistry()
        if len(self.tools) > 0:
            self.tools_registry.register_tools(self.tools)

    def agent_run(
        self, thread_messages: MutableSequence[MessageBase], **kwargs: Any
    ) -> Sequence[MessageBase]:
        """Execute agent's workflow entirely.

        The process of running the agent is separated in the following steps:
        1. Sent the messages to the model in `send_to_claude`. Retrieve model
            run response.
            
        2. Process the model response in `process_model_response`. If the
            response is a simple text message, `_current_run_messages` will be
            a list of `Message` objects.
            
        3. Process the tools in `process_tools`. Execute all required tools. 
            `_current_run_messages` will be a list containing `ToolUseMessage` 
            objects with the `tool_outputs` field populated.
    
        4. If there is any `ToolUseMessage` in the `_current_run_messages`, 
            call `agent_run` the agent again with the updated messages in order
             to get a new response from the model based on the tools outputs.

        Args:
        ----
            thread_messages: List of messages forming the conversation thread.
            **kwargs: Additional arguments for processing the response.

        Returns:
        -------
            Messages: List of messages generated by the agent. Including tool uses.

        Examples:
            >>> agent = Agent()
            >>> agent.agent_run([Message(content="Hello",type="text", role="user)])
            [Message(content="This is a Response", role="ai", type="text")]

        """
        if not kwargs.get("calling_from_inside_agent_run"):
            self._current_run_messages = []

        if self.verbose:
            logger.debug(f"Running agent with messages: {thread_messages}")

        response = self.send_to_claude(thread_messages, **kwargs)

        logger.debug(f"------------------\n{response}\n------------------")

        run_messages = self.process_model_response(response, **kwargs)
        self._current_run_messages.extend(run_messages)

        thread_messages.extend(run_messages)

        if any(
            isinstance(message, ToolUseMessage) for message in run_messages
        ):
            kwargs["calling_from_inside_agent_run"] = True
            logger.debug(
                "Tools were used. Feeding agent with results"
            ) if self.verbose else None
            self.agent_run(thread_messages, **kwargs)

        return self._current_run_messages

    def send_to_claude(
        self, thread_messages: MutableSequence[MessageBase], **kwargs: Any
    ) -> AnthropicMessage:
        """Send thread's messages to the model and return the raw response.

        Args:
        ----
            thread_messages: List of messages forming the conversation thread.
            **kwargs: Additional arguments for the serialization process.

        Returns:
        -------
            AnthropicMessage: The response message from the Claude model.

        """
        serialized_messages = self.messages_serializer(
            thread_messages,
            **{"roles_mapping": ModelMessageRoles.get_role_mapping()},
        )
        if self.verbose:
            logger.debug(f"Serialized messages: {serialized_messages}")

        if self.system_message_selector == "first":
            system_message = next(
                (
                    message
                    for message in thread_messages
                    if message.role == MessageRole.SYSTEM
                ),
                None,
            )

        elif self.system_message_selector == "last":
            system_message = next(
                (
                    message
                    for message in reversed(thread_messages)
                    if message.role == MessageRole.SYSTEM
                ),
                None,
            )

        serialized_tools = [self.tools_serializer(tool) for tool in self.tools]
        logger.debug(
            f"Serialized tools: {serialized_tools}"
        ) if self.verbose else None

        if system_message and isinstance(system_message, Message):
            logger.debug(
                f"Calling agent with {self.system_message_selector} "
                "system prompt."
            ) if self.verbose else None

            response = client.messages.create(
                model=self.model,
                max_tokens=self.max_tokens,
                messages=serialized_messages,
                tools=serialized_tools,
                system=system_message.content,
            )

        else:
            logger.debug(
                "Calling agent without system prompt."
            ) if self.verbose else None

            response = client.messages.create(
                model=self.model,
                max_tokens=self.max_tokens,
                messages=serialized_messages,
                tools=serialized_tools,
            )

        return response

    def process_model_response(
        self, response: AnthropicMessage, **kwargs: Any
    ) -> Sequence[MessageBase]:
        """Process the model response."""
        stop_reason = response.stop_reason
        if stop_reason in ["end_turn", "max_tokens", "stop_sequence"]:
            logger.debug("stop reason doesn't require tools processing.")
            if stop_reason == "max_tokens":
                logger.warning("The response reached the max tokens limit.")

            first_message = response.content[0]
            if isinstance(first_message, AnthropicTextBlock):
                messages: List[MessageBase] = [
                    Message(
                        role=MessageRole.AI,
                        type=MessageType.TEXT,
                        content=first_message.text,
                    )
                ]
                return messages
            else:
                raise ValueError(
                    f"Unexpected response content: {first_message}"
                )

        elif stop_reason == "tool_use":
            logger.debug("The response requires tools processing.")
            tool_use_messages: List[ToolUseMessage] = []
            # if the content blocks are not ToolUseBlocks, ignore them
            for block in response.content:
                if block.type == "tool_use":
                    if isinstance(block.input, dict) or isinstance(
                        block.input, str
                    ):
                        block_input = block.input
                    else:
                        block_input = str(block.input)

                    tool_use_message = ToolUseMessage(
                        run_id=block.id,
                        name=block.name,
                        type=MessageType.TEXT,
                        role=MessageRole.TOOL_USE,
                        input_params_dict=block_input,
                    )
                    tool_use_messages.append(tool_use_message)
                    
                else:
                    logger.warning(
                        f"Ignoring '{block.type}' block since a tool"
                        "will be used."
                    ) if self.verbose else None
                    # raise ValueError(f"Unexpected tool use block: {block}")
            logger.debug(f"Passing to process_tools: {tool_use_messages}")
            tool_use_messages = self.process_tools(tool_use_messages, **kwargs)

            return tool_use_messages

        else:
            raise ValueError(f"Unexpected stop reason: {stop_reason}")

    def process_tools(
        self, tool_use_messages: List[ToolUseMessage], **kwargs: Any
    ) -> List[ToolUseMessage]:
        """Process the tools."""
        updated_tool_use_messages = []
        if self.tools_registry:
            for tool_message in tool_use_messages:
                if isinstance(tool_message.input_params_dict, str):
                    args_dict = {"response": tool_message}
                else:
                    args_dict = tool_message.input_params_dict

                try:
                    logger.debug(
                        f"Executing tool: '{tool_message.name}' "
                        f"with args: {args_dict}"
                    )
                    tool_response = self.tools_registry.execute_tool(
                        tool_message.name,
                        args_dict,
                        **kwargs,
                    )
                    logger.info(
                        f"Tool returned:\n{tool_response}"
                    ) if self.verbose else None

                    tool_message.tool_outputs = tool_response.content
                    if tool_response.external_fields:
                        tool_message.external_fields.update(
                            tool_response.external_fields
                        )
                    updated_tool_use_messages.append(tool_message)

                except Exception:
                    tool_message.is_error = True
                    logger.error(
                        f"Error executing tool: {tool_message.name}"
                        f"with args: {args_dict}"
                    )
                    raise ValueError(f"Error executing tool: {tool_message}")

        return updated_tool_use_messages



from enum import Enum
from typing import Any, Dict

from light_agents.core.logger_config import setup_logger
from light_agents.schemas import ToolBaseSchema, ToolUseMessage
from light_agents.serializers.tools.base_serializers import (
    python_type_to_json_type,
)

logger = setup_logger(__name__)


def claude_tool_calling_serializer(
    llm_function: ToolBaseSchema,
) -> Dict[str, Any]:
    """Serialize a ToolBaseSchema into Claude API format."""
    function_dict = llm_function.model_dump(
        exclude={"name", "description", "required", "json_response"}
    )

    claude_format: Dict[str, Any] = {
        "name": llm_function.name,
        "description": llm_function.description,
        "input_schema": {
            "type": "object",
            "properties": {},
        },
    }

    if llm_function.required and len(llm_function.required) > 0:
        claude_format["input_schema"]["required"] = llm_function.required

    for param in function_dict.keys():
        json_param_type = python_type_to_json_type(function_dict[param])
        claude_format["input_schema"]["properties"][param] = {
            "type": json_param_type,
            "description": llm_function.model_fields[param].description,
        }

        annotation = llm_function.model_fields[param].annotation
        if (
            annotation
            and isinstance(annotation, type)
            and issubclass(annotation, Enum)
        ):
            claude_format["input_schema"]["properties"][param]["enum"] = list(
                annotation.__members__.keys()
            )

    return claude_format


def claude_tool_response_serializer(
    tool: ToolUseMessage, **kwargs: Any
) -> list[Dict[str, Any]]:
    """Serialize a single tool response for the Claude API.

    The function creates both ```assistant``` tool request block and ```user```
    tool response block.

    ```json
    [
        {
            "role": "assistant",
            "content": {
                "type": "tool_use",
                "id": "toolu_XXXXXX",
                "name": "tool_name",
                "input": {"arg1": "value1", "arg2": "value2"}
            }

        },
        {
            "role": "user",
            "content": [
                {
                    "type": "tool_result",
                    "tool_use_id": "toolu_XXXXXX",
                    "content": "the content to be passed to LLM"
                }
            ]
        }
    ]
    ```
    """
    if not isinstance(tool, ToolUseMessage):
        raise ValueError("The tool must be a ToolUseMessage instance.")

    tool_output = tool.tool_outputs
    if not isinstance(tool_output, str):
        logger.warning(
            "The tool output is not a string. Converting to string."
        )
        tool_output = str(tool.tool_outputs)

    tool_input = tool.input_params_dict
    if not isinstance(tool_input, dict):
        logger.warning(
            "The tool input is not a dictionary. Converting to dictionary."
        )
        tool_input = {"input": str(tool.input_params_dict)}

    serialized_tool_calling_message = {
        "role": "assistant",
        "content": [
            {
                "type": "tool_use",
                "id": tool.run_id,
                "name": tool.name,
                "input": tool.input_params_dict,
            }
        ],
    }

    serialized_tool_response_message = {
        "role": "user",
        "content": [
            {
                "type": "tool_result",
                "tool_use_id": tool.run_id,
                "content": tool_output,
            }
        ],
    }

    return [serialized_tool_calling_message, serialized_tool_response_message]
